from pathlib import Path
import os
import traceback
from panther.plugins.services.iut.quic.picoquic_shadow.config_schema import (
    PicoquicShadowConfig,
)
from panther.plugins.plugin_loader import PluginLoader
from panther.plugins.services.iut.implementation_interface import IImplementationManager
from panther.plugins.protocols.config_schema import ProtocolConfig, RoleEnum


class PicoquicShadowServiceManager(IImplementationManager):
    """
    PicoquicShadowServiceManager is a service manager for handling Picoquic services.
    This class is responsible for initializing the service manager, generating various commands required for the service lifecycle, and preparing the service manager for use.
    Methods:
        __init__(self, service_config_to_test: PicoquicShadowConfig, service_type: str, protocol: ProtocolConfig, implementation_name: str):
            Initializes the PicoquicShadowServiceManager with the given configuration, service type, protocol, and implementation name.
        get_service_name(self) -> str:
            Returns the name of the service.
        generate_pre_compile_commands(self):
        generate_compile_commands(self):
        generate_pre_run_commands(self):
        generate_run_command(self):
        generate_post_run_commands(self):
        prepare(self, plugin_loader: Optional[PluginLoader] = None):
            Prepares the service manager for use.
        generate_deployment_commands(self) -> str:
    """

    def __init__(
        self,
        service_config_to_test: PicoquicShadowConfig,
        service_type: str,
        protocol: ProtocolConfig,
        implementation_name: str,
    ):
        super().__init__(
            service_config_to_test, service_type, protocol, implementation_name
        )
        self.logger.debug(
            f"Initializing Picoquic service manager for '{implementation_name}'"
        )
        self.logger.debug(
            f"Loaded Picoquic configuration: {self.service_config_to_test}"
        )
        self.initialize_commands()

    def get_service_name(self) -> str:
        return self.service_name

    def generate_pre_compile_commands(self):
        """
        Generates pre-compile commands.
        """
        return super().generate_pre_compile_commands() + []

    def generate_compile_commands(self):
        """
        Generates compile commands.
        """
        return super().generate_compile_commands() + []

    def generate_pre_run_commands(self):
        """
        Generates pre-run commands.
        """
        return super().generate_pre_run_commands() + []

    def generate_run_command(self):
        """
        Generates the run command.
        """
        cmd_args = self.generate_deployment_commands()
        return {
            "working_dir": self.working_dir,
            "command_binary": (
                self.service_config_to_test.implementation.version.server.binary.name
                if self.role == RoleEnum.server
                else self.service_config_to_test.implementation.version.client.binary.name
            ),
            "command_args": cmd_args,
            "timeout": self.service_config_to_test.timeout,
            "command_env": {},
        }

    def generate_post_run_commands(self):
        """
        Generates post-run commands.
        """
        return super().generate_post_run_commands() + [
            "cp /opt/picoquic/picoquicdemo /app/logs/picoquicdemo;"
        ]

    def prepare(self, plugin_loader: PluginLoader | None = None):
        """
        Prepare the service manager for use.
        """
        self.logger.debug("Preparing Picoquic service manager...")
        plugin_loader.build_docker_image_from_path(
            Path(
                os.path.join(
                    os.getcwd(),
                    "panther",
                    "plugins",
                    "services",
                    "Dockerfile",
                )
            ),
            "panther_base",
            "service",
        )
        plugin_loader.build_docker_image(
            self.get_implementation_name(),
            self.service_config_to_test.implementation.version,
        )

    def generate_deployment_commands(self) -> str:
        """
        Generates deployment commands and collects volume mappings based on service parameters.

        :param service_params: Parameters specific to the service.
        :param environment: The environment in which the services are being deployed.
        :return: A dictionary with service name as key and a dictionary containing command and volumes.
        """
        self.logger.debug(
            f"Generating deployment commands for service: {self.service_name} with service parameters: {self.service_config_to_test}"
        )
        # Create the command list

        self.logger.debug(f"Role: {self.role}, Version: {self.service_version}")

        # Determine if network interface parameters should be included based on environment
        # TODO
        # include_interface = environment not in ["docker_compose"]
        include_interface = True

        # Build parameters for the command template
        # TODO ensure that the parameters are correctly set
        if self.role == RoleEnum.server:
            params = self.service_config_to_test.implementation.version.server
        # For the client, include target and message if available
        elif self.role == RoleEnum.client:
            params = self.service_config_to_test.implementation.version.client

        params["target"] = self.service_config_to_test.protocol.target

        self.logger.debug(f"Parameters for command template: {params}")
        self.logger.debug(f"Role: {self.role}")
        self.working_dir = params["binary"]["dir"]
        # Conditionally include network interface parameters
        if not include_interface:
            params["network"].pop("interface", None)

        # Render the appropriate template
        try:
            template_name = f"{str(self.role.name)}_command.jinja"
            return super().render_commands(params, template_name)

        except Exception as e:
            self.logger.error(
                f"Failed to render command for service '{self.service_config_to_test.name}': {e}\n{traceback.format_exc()}"
            )
            raise e



from abc import ABC

from panther.config.config_experiment_schema import ServiceConfig
from panther.plugins.services.services_interface import IServiceManager


class ITesterManager(IServiceManager, ABC):

    def __init__(
        self,
        service_config_to_test: ServiceConfig,
        service_type: str,
        protocol: str,
        implementation_name: str,
    ):
        super().__init__(
            service_config_to_test, service_type, protocol, implementation_name
        )




from panther.core.results.result_handler import ResultHandler

class ParserHandler(ResultHandler):
    """
    A handler class for parsing experiment results according to the environment 
    and the implementation under test.
    Attributes:
        parser: An instance of a parser that will be used to parse the request.
    Methods:
        handle(request):
            Parses the given request using the parser instance.
    """
    
    def __init__(self, parser) -> None:
        self.parser = parser
        
    def handle(self, request) -> None:
        self.parser.parse(request)


from abc import ABC

from omegaconf import OmegaConf

from panther.core.observer.event_manager import EventManager
from panther.config.config_experiment_schema import TestConfig
from panther.config.config_global_schema import GlobalConfig
from panther.plugins.environments.execution_environment.strace.config_schema import (
    StraceConfig,
)
from panther.plugins.environments.execution_environment.execution_environment_interface import (
    IExecutionEnvironment,
)
from panther.plugins.plugin_loader import PluginLoader
from panther.plugins.services.services_interface import IServiceManager


class StraceEnvironment(IExecutionEnvironment, ABC):
    """
    StraceEnvironment is a class that sets up and manages an execution environment using strace for system call tracing.

    Attributes:
        global_config (GlobalConfig): The global configuration for the environment.
        env_config_to_test (StraceConfig): The specific configuration for the strace environment to test.
        services_managers (list[IServiceManager]): List of service managers to handle services within the environment.
        test_config (TestConfig): Configuration for the test being executed.
        plugin_loader (PluginLoader): Loader for plugins used in the environment.

    Methods:
        __init__(env_config_to_test: StraceConfig, output_dir: str, env_type: str, env_sub_type: str, event_manager: EventManager):
            Initializes the StraceEnvironment with the given configuration and parameters.

        setup_environment(services_managers: list[IServiceManager], test_config: TestConfig, global_config: GlobalConfig, timestamp: str, plugin_loader: PluginLoader):
            Sets up the environment with the provided service managers, test configuration, global configuration, and plugin loader.

        to_command(pid: int | None = None) -> str:
            Generates the strace command for execution. Optionally attaches to a specific process ID.

        __repr__() -> str:
            Returns a string representation of the StraceEnvironment instance.
    """
    # TODO enforce config in environment
    def __init__(
        self,
        env_config_to_test: StraceConfig,
        output_dir: str,
        env_type: str,
        env_sub_type: str,
        event_manager: EventManager,
    ):
        super().__init__(
            env_config_to_test, output_dir, env_type, env_sub_type, event_manager
        )
        self.global_config = None
        self.env_config_to_test = env_config_to_test

    def setup_environment(
        self,
        services_managers: list[IServiceManager],
        test_config: TestConfig,
        global_config: GlobalConfig,
        timestamp: str,
        plugin_loader: PluginLoader,
    ):
        self.services_managers: list[IServiceManager] = services_managers
        self.test_config = test_config
        self.plugin_loader = plugin_loader
        self.global_config = global_config
        self.logger.debug("Setup environment with:")
        self.logger.debug(f"Services config: {self.env_config_to_test}")
        for service in self.services_managers:
            self.logger.debug(f"Service cmds: {service.run_cmd}")
            service.run_cmd["pre_run_cmds"] = service.run_cmd["pre_run_cmds"] + [
                self.to_command()
            ]
            self.logger.debug(f"Service cmds: {service.run_cmd}")

        self.logger.debug(f"Test Config: {OmegaConf.to_yaml(self.test_config)}")
        self.logger.debug(f"Global Config: {OmegaConf.to_yaml(self.global_config)}")

    def to_command(self, pid: int | None = None) -> str:
        """
        Generate the strace command for execution.
        :param pid: Optional process ID to attach to.
        :return: Strace command as a string.
        """
        self.env_config_to_test = StraceConfig()
        excluded = ",".join(
            f"{syscall}" for syscall in self.env_config_to_test.excluded_syscalls
        )
        command = [
            self.env_config_to_test.strace_binary,
            "-k",
        ]  # Include kernel stack if enabled
        command.append(f'-e trace="!{excluded}"')  # Exclude specified syscalls
        if pid:
            command.extend(["-p", str(pid)])
        # if self.env_config_to_test.trace_network_syscalls:
        #     command.append("-e trace=network")  # Include network-related syscalls
        # if self.env_config_to_test.additional_parameters:
        #     command.extend(self.env_config_to_test.additional_parameters)
        # command.append(f"-o {self.env_config_to_test.output_file}")
        return " ".join(command)

    def __repr__(self):
        return (
            f"StraceEnvironment(env_config_to_test={self.env_config_to_test}, "
            f"output_dir={self.output_dir}, event_manager={self.event_manager}, "
            f"services_managers={self.services_managers}, test_config={self.test_config})"
        )



from abc import abstractmethod
import os
from panther.core.observer.event_manager import EventManager
from panther.plugins.environments.config_schema import EnvironmentConfig
from panther.plugins.plugin_interface import IPlugin


class IEnvironmentPlugin(IPlugin):
    """
    IEnvironmentPlugin is an abstract base class that defines the interface for environment plugins.

    Attributes:
        templates_dir (str): Directory path for templates specific to the environment type and subtype.
        output_dir (str): Directory path for output files.
        env_type (str): Type of the environment.
        env_sub_type (str): Subtype of the environment.
        log_dirs (str): Directory path for log files.
        plugin_loader: Loader for the plugin (initially set to None).
        env_config_to_test (EnvironmentConfig): Configuration of the environment to be tested.
        event_manager (EventManager): Manager for handling events.

    Methods:
        is_network_environment():
            Abstract method. Returns True if the plugin is a network environment.

        setup_environment():
            Abstract method. Sets up the required environment before running experiments.

        teardown_environment():
            Abstract method. Tears down the environment after experiments are completed.
    """

    def __init__(
        self,
        env_config_to_test: EnvironmentConfig,
        output_dir: str,
        env_type: str,
        env_sub_type: str,
        event_manager: EventManager,
    ):
        super().__init__()
        self.templates_dir: str = (
            f"panther/plugins/environments/{env_type}/{env_sub_type}/templates"
        )
        self.output_dir = output_dir
        self.env_type = env_type
        self.env_sub_type = env_sub_type
        self.log_dirs = os.path.join(self.output_dir, "logs")
        self.plugin_loader = None
        self.env_config_to_test = env_config_to_test
        self.event_manager = event_manager

    @abstractmethod
    def is_network_environment(self):
        """
        Returns True if the plugin is a network environment.
        """
        pass

    @abstractmethod
    def setup_environment(self):
        """
        Sets up the required environment before running experiments.
        """
        pass

    @abstractmethod
    def teardown_environment(self):
        """
        Tears down the environment after experiments are completed.
        """
        pass



import importlib
import logging
import os
from pathlib import Path
import shutil
from omegaconf import DictConfig, OmegaConf, ValidationError, ListConfig
import yaml
from panther.config.config_global_schema import (
    DockerConfig,
    GlobalConfig,
    LoggingConfig,
    PathsConfig,
)
from panther.config.config_experiment_schema import (
    ExperimentConfig,
    ServiceConfig,
    TestConfig,
)
from panther.plugins.plugin_loader import PluginLoader


class ConfigLoader:
    def __init__(
        self,
        experiment_file: str,
        output_dir: str | None = None,
        exec_env_dir: str | None = "",
        net_env_dir: str | None = "",
        iut_dir: str | None = "",
        testers_dir: str | None = "",
    ):

        self.experiment_file = experiment_file
        self.output_dir = output_dir

        self.exec_env_dir = exec_env_dir
        self.net_env_dir = net_env_dir
        self.iut_dir = iut_dir
        self.testers_dir = testers_dir

        self.logger = logging.getLogger("ConfigLoader")
        self.global_config = None

    def construct_global_config(self, loaded_config: DictConfig) -> GlobalConfig:
        """_summary_

        Args:
            loaded_config (DictConfig): _description_

        Returns:
            ExperimentConfig: _description_
        """
        # Construct logging configuration
        logging_config = LoggingConfig(
            level=loaded_config["logging"]["level"],
            format=loaded_config["logging"]["format"],
        )
        OmegaConf.merge(LoggingConfig, logging_config)

        # Construct paths configuration
        paths_config = PathsConfig(
            output_dir=(
                loaded_config["paths"]["output_dir"]
                if not self.output_dir
                else self.output_dir
            ),
            log_dir=loaded_config["paths"]["log_dir"],
            config_dir=loaded_config["paths"]["config_dir"],
            plugin_dir=loaded_config["paths"]["plugin_dir"],
        )
        OmegaConf.merge(PathsConfig, paths_config)

        # TODO: Not used for now
        # optional_paths_config = AdditionalPathsConfig(
        #     exec_env_dir=self.exec_env_dir,
        #     net_env_dir=self.net_env_dir,
        #     iut_dir=self.iut_dir,
        #     testers_dir=self.testers_dir,
        # )

        self.add_plugin_execution_environment()

        self.add_plugin_network_environment()

        self.add_plugin_iut_service()

        self.add_plugin_tester_service()

        # Construct Docker configuration
        docker_config = DockerConfig(
            build_docker_image=loaded_config["docker"]["build_docker_image"]
        )
        OmegaConf.merge(DockerConfig, docker_config)

        global_config = GlobalConfig(
            logging=logging_config,
            paths=paths_config,
            # optional_paths=optional_paths_config,
            docker=docker_config,
        )
        OmegaConf.merge(GlobalConfig, global_config)
        self.global_config = global_config
        return global_config

    def add_plugin_tester_service(self):
        """
        Adds a plugin tester service by copying tester files from the specified directory.

        This method checks if the `testers_dir` attribute is set and not empty. If so, it converts
        `testers_dir` to a `Path` object and constructs the target directory path for the testers.
        It then calls the `copy_plugin_files` method to copy the tester files to the target directory.

        Raises:
            Exception: If the `testers_dir` attribute is not set or is empty.
        """
        if self.testers_dir and self.testers_dir != "":
            print(f"Copying testers from {self.testers_dir}")
            self.testers_dir = Path(self.testers_dir)
            testers_target_dir = os.path.join(
                "panther", "plugins", "services", "testers", self.testers_dir.name
            )
            self.copy_plugin_files(testers_target_dir)

    def copy_plugin_files(self, testers_target_dir):
        """
        Copies plugin files from the source directory to the target directory.

        This method checks if the target directory exists, and if not, it creates it.
        It then iterates through all items in the source directory (self.testers_dir),
        and copies each item to the target directory. If an item is a directory, it
        recursively copies the entire directory. If an item is a file, it copies the file.

        Parameters:
        testers_target_dir (str): The path to the target directory where plugin files
                      should be copied.

        Raises:
            OSError: If the source directory does not exist or if there is an error during
             the copying process.
        """
        if not os.path.exists(testers_target_dir):
            os.makedirs(testers_target_dir)
        for item in os.listdir(self.testers_dir):
            print(f"Copying {item} from {self.testers_dir} to {testers_target_dir}")
            s = os.path.join(self.testers_dir, item)
            d = os.path.join(testers_target_dir, item)
            if os.path.isdir(s):
                if os.path.exists(d):
                    shutil.rmtree(d)
                shutil.copytree(s, d)
            else:
                shutil.copy2(s, d)

    def remove_plugin_tester_service(self):
        """
        Removes the plugin tester service directory if it exists.

        This method checks if the `testers_dir` attribute is set and not empty.
        If so, it constructs the path to the target directory within the
        "panther/plugins/services/testers" directory. If the target directory
        exists, it removes the directory and all its contents.

        Attributes:
            testers_dir (str): The directory path of the testers to be removed.

        Side Effects:
            Deletes the directory specified by `testers_dir` and all its contents.

        Prints:
            A message indicating the removal of the testers directory.
        """
        if self.testers_dir and self.testers_dir != "":
            testers_target_dir = os.path.join(
                "panther", "plugins", "services", "testers", Path(self.testers_dir).name
            )
            if os.path.exists(testers_target_dir):
                print(f"Removing testers from {testers_target_dir}")
                shutil.rmtree(testers_target_dir)

    def add_plugin_iut_service(self):
        if self.iut_dir and self.iut_dir != "":
            print(f"Copying IUT from {self.iut_dir}")
            self.iut_dir = Path(self.iut_dir)
            iut_target_dir = os.path.join(
                "panther", "plugins", "services", "iut", self.iut_dir.name
            )
            self.copy_plugin_files(iut_target_dir)

    def remove_plugin_iut_service(self):
        if self.iut_dir and self.iut_dir != "":
            iut_target_dir = os.path.join(
                "panther", "plugins", "services", "iut", Path(self.iut_dir).name
            )
            if os.path.exists(iut_target_dir):
                print(f"Removing IUT from {iut_target_dir}")
                shutil.rmtree(iut_target_dir)

    def add_plugin_network_environment(self):
        if self.net_env_dir and self.net_env_dir != "":
            # TODO improve this
            print(f"Copying network environment from {self.net_env_dir}")
            self.net_env_dir = Path(self.net_env_dir)
            net_env_target_dir = os.path.join(
                "panther",
                "plugins",
                "environments",
                "network_environment",
                self.net_env_dir.name,
            )
            self.copy_plugin_files(net_env_target_dir)

    def remove_plugin_network_environment(self):
        if self.net_env_dir and self.net_env_dir != "":
            net_env_target_dir = os.path.join(
                "panther",
                "plugins",
                "environments",
                "network_environment",
                Path(self.net_env_dir).name,
            )
            if os.path.exists(net_env_target_dir):
                print(f"Removing network environment from {net_env_target_dir}")
                shutil.rmtree(net_env_target_dir)

    def add_plugin_execution_environment(self):
        if self.exec_env_dir and self.exec_env_dir != "":
            # TODO improve this
            print(f"Copying execution environment from {self.exec_env_dir}")
            self.exec_env_dir = Path(self.exec_env_dir)
            exec_env_target_dir = os.path.join(
                "panther",
                "plugins",
                "environments",
                "execution_environment",
                self.exec_env_dir.name,
            )
            self.copy_plugin_files(exec_env_target_dir)

    def remove_plugin_execution_environment(self):
        if self.exec_env_dir and self.exec_env_dir != "":
            exec_env_target_dir = os.path.join(
                "panther",
                "plugins",
                "environments",
                "execution_environment",
                Path(self.exec_env_dir).name,
            )
            if os.path.exists(exec_env_target_dir):
                print(f"Removing execution environment from {exec_env_target_dir}")
                shutil.rmtree(exec_env_target_dir)

    def cleanup(self):
        self.remove_plugin_execution_environment()
        self.remove_plugin_network_environment()
        self.remove_plugin_iut_service()
        self.remove_plugin_tester_service()

    def validate_plugin_config(
        self, plugin_type: str, plugin_name: str, plugin_config: DictConfig
    ):
        """
        Validate plugin-specific configuration against its schema.

        :param plugin_type: The plugin type (e.g., "network_environment").
        :param plugin_name: The plugin name (e.g., "shadow_ns").
        :param plugin_config: The plugin configuration to validate.
        :return: Validated plugin configuration.
        :raises ValidationError: If the configuration does not conform to the schema.
        """
        self.logger.debug(
            f"Validating plugin configuration for {plugin_type}/{plugin_name} with {plugin_config}"
        )
        plugin_schema_class = self.load_plugin_schema(plugin_type, plugin_name)
        self.logger.debug(f"Plugin schema class: {plugin_schema_class}")
        structured_schema = OmegaConf.structured(plugin_schema_class)
        self.logger.debug(f"Structured schema: {structured_schema}")
        try:
            return OmegaConf.merge(structured_schema, plugin_config)
        except ValidationError as e:
            raise ValidationError(
                f"Plugin configuration validation failed for {plugin_type}/{plugin_name}: {e}"
            )

    def construct_experiment_config(
        self, loaded_config: DictConfig
    ) -> ExperimentConfig:
        """
        Manually construct an ExperimentConfig object from a loaded configuration.

        :param loaded_config: The loaded configuration dictionary (DictConfig or dict).
        :return: An ExperimentConfig object.
        """
        self.logger.info(
            f"Constructing experiment configuration with global configuration - {self.global_config}"
        )
        # Construct tests
        tests: list[TestConfig] = []
        for test_data in loaded_config["tests"]:
            # Construct network environment configuration
            network_env = test_data["network_environment"]
            self.logger.debug(f"Network environment: {network_env}")
            validated_network_env = self.validate_plugin_config(
                "network_environment", network_env["type"], network_env
            )
            self.logger.debug(
                f"Network environment after validation: {validated_network_env}"
            )

            exec_envs = test_data.get("execution_environment", [])
            self.logger.debug(f"Execution environment: {exec_envs}")
            for exec_env in exec_envs:
                validated_network_env = self.validate_plugin_config(
                    "execution_environment", exec_env["type"], exec_env
                )
            self.logger.debug(
                f"Network environment after validation: {validated_network_env}"
            )

            # Construct services for this test
            services: dict[str, ServiceConfig] = {}
            for service_name, service_data in test_data["services"].items():
                protocol = self.load_and_validate_protocol_config(
                    service_data
                )  # Resolve protocol subclass
                self.logger.debug(f"Protocol: {protocol}")
                implementation = self.load_and_validate_implementation_config(
                    service_data
                )
                self.logger.debug(f"Implementation: {implementation}")
                service = ServiceConfig(
                    name=service_data["name"],
                    timeout=service_data.get("timeout", 100),
                    implementation=implementation,
                    protocol=protocol,
                    ports=service_data.get("ports", []),
                    generate_new_certificates=service_data.get(
                        "generate_new_certificates", False
                    ),
                )
                OmegaConf.merge(ServiceConfig, service)
                services[service_name] = service

            # ##########################################################################################################
            # Construct the test configuration:
            # NOTE: We do not validate with merge here, as the schema is not fully compatible with OmegaConf
            # It is because NetworkEnvironmentConfig is a dataclass, and OmegaConf does not support nested dataclasses
            # For example, let
            # class NetworkEnvironmentConfig:
            #     type: str
            # And:
            # class DockerComposeConfig(NetworkEnvironmentConfig):
            #     type: str    = "docker_compose"
            #     version: str = "3.8"#omegaconf.errors.ConfigKeyError: Key 'version' not in 'NetworkEnvironmentConfig'
            #     network_name: str = "default_network"
            #     service_prefix: Optional[str] = None  # Optional prefix for service names
            #     volumes: List[str] = field(default_factory=list)  # List of volume mounts
            #     environment: Dict[str, str] = field(default_factory=dict)  # Environment variables
            # Thus we prelinarily validate the network environment configuration,
            # It will thus ignore non defined fields in the schema -> not ideal for validation
            # We will need to find a way to validate nested dataclasses with OmegaConf
            # TODO if tests name is undefined, use the service name + other parameters
            # TODO if we use the validated version -> bugs (but it should be enough to validate format)
            # ##########################################################################################################

            test = TestConfig(
                name=test_data["name"],
                description=test_data["description"],
                network_environment=network_env,
                execution_environments=test_data.get("execution_environment", []),
                iterations=test_data["iterations"],
                services=services,
                steps=test_data.get("steps"),
                assertions=test_data.get("assertions"),
            )
            tests.append(test)

        # Construct the ExperimentConfig
        experiment_config = ExperimentConfig(
            tests=tests,
        )
        return experiment_config

    def load_and_validate_experiment_config(self) -> ExperimentConfig:
        """
        Load and validate the entire experiment configuration, including plugin-specific validation.

        :return: A validated experiment configuration.
        """
        experiment_config_path = self.experiment_file
        if not os.path.exists(experiment_config_path):
            raise FileNotFoundError(
                f"Experiment configuration file '{experiment_config_path}' not found."
            )

        try:
            # Load the YAML configuration
            loaded_config = OmegaConf.load(experiment_config_path)
            self.logger.debug(f"Loaded experiment config: {loaded_config}")
            self.logger.debug(OmegaConf.to_yaml(OmegaConf.structured(ExperimentConfig)))
            experiment_config = self.construct_experiment_config(loaded_config)
            self.logger.debug(
                f"Constructed experiment config: {OmegaConf.to_yaml(experiment_config)}"
            )
            self.logger.info("Experiment configuration successfully validated.")
            return experiment_config
        except ValidationError as e:
            self.logger.error(f"Configuration validation failed: {e}")
            raise
        except yaml.parser.ParserError as e:
            self.logger.error(f"YAML parsing error: {e}")
            raise
        except Exception as e:
            self.logger.error(f"Unexpected error during configuration loading: {e}")
            raise

    def load_and_validate_global_config(self) -> GlobalConfig:
        """
        Load and validate the entire experiment configuration, including plugin-specific validation.
        Note no logger is used here.

        :return: A validated experiment configuration.
        """
        experiment_config_path = self.experiment_file
        if not os.path.exists(experiment_config_path):
            raise FileNotFoundError(
                f"Experiment configuration file '{experiment_config_path}' not found."
            )

        try:
            # Load the YAML configuration
            loaded_config = OmegaConf.load(experiment_config_path)
            print(f"Loaded experiment config: {loaded_config}")
            print(OmegaConf.to_yaml(OmegaConf.structured(ExperimentConfig)))
            global_config = self.construct_global_config(loaded_config)
            print(f"Constructed global config: {OmegaConf.to_yaml(global_config)}")
            print("Experiment configuration successfully validated.")
            return global_config
        except ValidationError as e:
            print(f"Configuration validation failed: {e}")
            raise
        except yaml.parser.ParserError as e:
            print(f"YAML parsing error: {e}")
            raise
        except Exception as e:
            print(f"Unexpected error during configuration loading: {e}")
            raise

    @staticmethod
    def load_plugin_schema(plugin_type: str, plugin_name: str):
        """
        Dynamically load a plugin schema based on its type and name.

        :param plugin_type: The plugin type (e.g., "network_environment").
        :param plugin_name: The plugin name (e.g., "shadow_ns").
        :return: The plugin's schema module.
        :raises ImportError: If the schema module cannot be found.
        """
        plugin_module_path = (
            f"panther.plugins.environments.{plugin_type}.{plugin_name}.config_schema"
        )
        try:
            class_name = PluginLoader.get_class_name(plugin_name)
            plugin_module = importlib.import_module(plugin_module_path)
            config_class = getattr(plugin_module, class_name)
            return config_class  # Assume PluginConfig is the schema class
        except ImportError:
            raise ImportError(f"Plugin schema '{plugin_module_path}' not found.")
        except AttributeError:
            raise ImportError(
                f"Plugin schema '{plugin_module_path}' does not define a 'PluginConfig' class."
            )

    def load_and_validate_protocol_config(
        self, implementation: ServiceConfig
    ) -> ListConfig | DictConfig:
        """
        Dynamically loads the appropriate implementation configuration class.

        :param implementation: A dictionary containing `name` and other fields.
        :return: An instance of the dynamically loaded configuration class.
        """
        self.logger.debug(f"Service: {implementation}")
        protocol = implementation.protocol.name
        if hasattr(implementation.protocol, "protocol_type"):
            protocol_type = implementation.protocol.protocol_type
        else:
            protocol_type = "client_server"  # TODO: Default to client-server for now
            # Assuming schema files are in plugins
        module_path = (
            f"panther.plugins.protocols.{protocol_type}.{protocol}.config_schema"
        )
        try:
            # Import the module and dynamically get the class
            schema_module = importlib.import_module(module_path)
            config_class = getattr(schema_module, f"{protocol.capitalize()}Config")
            self.logger.debug(
                f"Protocol: {protocol} - {implementation['protocol']} - {config_class}"
            )
            protocol_instance = config_class(**implementation.protocol)
            return OmegaConf.merge(config_class, protocol_instance)
        except (ImportError, AttributeError) as e:
            raise ValueError(f"Failed to load protocol config for '{protocol}': {e}")

    def load_and_validate_implementation_config(
        self, implementation: dict
    ) -> ListConfig | DictConfig:
        """
        Dynamically loads the appropriate implementation configuration class.
        """
        self.logger.debug(f"Implementation: {implementation}")
        name = implementation["implementation"]["name"]
        implem_type = implementation["implementation"]["type"]
        protocol = implementation["protocol"]["name"]
        protocol_version = implementation["protocol"]["version"]
        if implem_type == "iut":
            # Assuming schema files are in plugins
            module_path = f"panther.plugins.services.{implem_type}.{protocol}.{name}.config_schema"
        else:
            module_path = f"panther.plugins.services.{implem_type}.{name}.config_schema"

        self.logger.debug(f"Module path: {module_path}")
        try:
            # Import the module and dynamically get the class
            schema_module = importlib.import_module(module_path)
            class_name = PluginLoader.get_class_name(name)
            config_class = getattr(schema_module, class_name)
            self.logger.debug(
                f"Implementation: {name} - {implementation['implementation']} - {config_class}"
            )

            # Load the version configuration
            version_class_name = PluginLoader.get_class_name(name, "Version")
            version_config_class = getattr(schema_module, version_class_name)
            if implem_type == "iut":
                version_configs_dir = module_path.replace(".", "/").replace(
                    "/config_schema", "/version_configs/"
                )
            else:
                version_configs_dir = module_path.replace(".", "/").replace(
                    "/config_schema", f"/version_configs/{protocol}/"
                )

            version_path = os.path.join(version_configs_dir, f"{protocol_version}.yaml")
            if not os.path.exists(version_path):
                raise ValueError(
                    f"Version configuration file {version_path} not found."
                )
            raw_version_config = OmegaConf.load(version_path)
            self.logger.debug(
                f"Version config: {raw_version_config} - {version_config_class}"
            )
            protocol_version = OmegaConf.to_object(
                OmegaConf.merge(
                    OmegaConf.structured(version_config_class), raw_version_config
                )
            )

            implementation_instance = config_class(**implementation["implementation"])
            implementation_instance.version = protocol_version

            return OmegaConf.merge(config_class, implementation_instance)
        except (ImportError, AttributeError) as e:
            raise ValueError(f"Failed to load implementation config for '{name}': {e}")

    def get_all_exec_env_classes(self):
        """
        Get all execution environment classes from the plugin directory.

        :return: A list of execution environment classes.
        """
        exec_env_classes = []
        exec_env_dir = (
            Path(self.global_config.paths.plugin_dir)
            / "environments"
            / "execution_environment"
        )
        self.logger.debug(
            f"Searching for execution environment classes in {exec_env_dir}"
        )
        for plugin_dir in exec_env_dir.iterdir():
            if plugin_dir.is_dir():
                plugin_file = plugin_dir / f"{plugin_dir.name}.py"
                if plugin_file.exists():
                    self.logger.debug(
                        f"Found execution environment class: {plugin_dir.name}"
                    )
                    exec_env_classes.append(
                        PluginLoader.get_class_name(plugin_dir.name, "Config")
                    )
                else:
                    self.logger.debug(
                        f"No execution environment class found in {plugin_dir}"
                    )
        self.logger.debug(
            f"Total execution environment classes found: {len(exec_env_classes)}"
        )
        return exec_env_classes

    def get_all_net_env_classes(self):
        """
        Get all network environment classes from the plugin directory.

        :return: A list of network environment classes.
        """
        exec_env_classes = []
        exec_env_dir = (
            Path(self.global_config.paths.plugin_dir)
            / "environments"
            / "network_environment"
        )
        self.logger.debug(
            f"Searching for network environment classes in {exec_env_dir}"
        )
        for plugin_dir in exec_env_dir.iterdir():
            if plugin_dir.is_dir():
                plugin_file = plugin_dir / f"{plugin_dir.name}.py"
                if plugin_file.exists():
                    self.logger.debug(
                        f"Found network environment class: {plugin_dir.name}"
                    )
                    exec_env_classes.append(
                        PluginLoader.get_class_name(plugin_dir.name, "Config")
                    )
                else:
                    self.logger.debug(
                        f"No network environment class found in {plugin_dir}"
                    )
        self.logger.debug(
            f"Total network environment classes found: {len(exec_env_classes)}"
        )
        return exec_env_classes

    def get_all_protocol_classes(self):
        """
        Get all protocol classes from the plugin directory.

        :return: A list of protocol classes.
        """
        protocol_classes = []
        protocol_dir = Path(self.global_config.paths.plugin_dir) / "protocols"
        self.logger.debug(f"Searching for protocol classes in {protocol_dir}")
        for protocol_type_dir in protocol_dir.iterdir():
            if protocol_type_dir.is_dir():
                for protocol_dir in protocol_type_dir.iterdir():
                    if protocol_dir.is_dir():
                        protocol_file = protocol_dir / f"{protocol_dir.name}.py"
                        if protocol_file.exists():
                            self.logger.debug(
                                f"Found protocol class: {protocol_dir.name}"
                            )
                            protocol_classes.append(
                                PluginLoader.get_class_name(protocol_dir.name, "Config")
                            )
                        else:
                            self.logger.debug(
                                f"No protocol class found in {protocol_dir}"
                            )
        self.logger.debug(f"Total protocol classes found: {len(protocol_classes)}")
        return protocol_classes

    def get_all_iut_classes(self):
        """
        Get all IUT classes from the plugin directory and return a dictionary with protocols as keys and list of implementations as values.

        :return: A dictionary with protocols as keys and list of IUT classes as values.
        """
        iut_classes = {}
        iut_dir = Path(self.global_config.paths.plugin_dir) / "services" / "iut"
        self.logger.debug(f"Searching for IUT classes in {iut_dir}")
        for protocol_dir in iut_dir.iterdir():
            if protocol_dir.is_dir():
                protocol_name = protocol_dir.name
                iut_classes[protocol_name] = []
                for plugin_dir in protocol_dir.iterdir():
                    if plugin_dir.is_dir() and not plugin_dir.name.startswith("__"):
                        plugin_file = plugin_dir / f"{plugin_dir.name}.py"
                        if plugin_file.exists():
                            self.logger.debug(f"Found IUT class: {plugin_dir.name}")
                            iut_classes[protocol_name].append(
                                PluginLoader.get_class_name(plugin_dir.name, "Config")
                            )
                        else:
                            self.logger.debug(f"No IUT class found in {plugin_dir}")
        self.logger.debug(f"Total IUT classes found: {iut_classes}")
        return iut_classes

    def get_all_tester_classes(self):
        """
        Get all tester classes from the plugin directory and return a list of tester classes.

        :return: A list of tester classes.
        """
        tester_classes = []
        tester_dir = Path(self.global_config.paths.plugin_dir) / "services" / "testers"
        self.logger.debug(f"Searching for tester classes in {tester_dir}")
        for plugin_dir in tester_dir.iterdir():
            if plugin_dir.is_dir():
                plugin_file = plugin_dir / f"{plugin_dir.name}.py"
                if plugin_file.exists():
                    self.logger.debug(f"Found tester class: {plugin_dir.name}")
                    tester_classes.append(
                        PluginLoader.get_class_name(plugin_dir.name, "Config")
                    )
                else:
                    self.logger.debug(f"No tester class found in {plugin_dir}")
        self.logger.debug(f"Total tester classes found: {len(tester_classes)}")
        return tester_classes

    def load_all_plugins(self) -> dict[str, list[str]]:
        """
        Load all plugins and return them in an ordered dictionary.

        :return: An ordered dictionary with plugin types as keys and list of plugin names as values.
        """
        plugin_types = {
            "execution_environment": "environments/execution_environment",
            "network_environment": "environments/network_environment",
            "protocols": "protocols",
            "iut": "services/iut",
            "testers": "services/testers",
        }
        all_plugins = {}

        for plugin_type, plugin_path in plugin_types.items():
            plugin_dir = Path(self.global_config.paths.plugin_dir) / plugin_path
            self.logger.debug(f"Searching for plugins in {plugin_dir}")
            plugins = []
            if plugin_type == "iut":
                for protocol_dir in plugin_dir.iterdir():
                    if protocol_dir.is_dir() and not protocol_dir.name.startswith("__"):
                        for sub_dir in protocol_dir.iterdir():
                            if sub_dir.is_dir() and not sub_dir.name.startswith("__"):
                                plugin_file = sub_dir / f"{sub_dir.name}.py"
                                if plugin_file.exists():
                                    self.logger.debug(f"Found plugin: {sub_dir.name}")
                                    plugins.append(
                                        f"{protocol_dir.name}/{sub_dir.name}"
                                    )
                                    try:
                                        importlib.import_module(
                                            f"panther.plugins.{plugin_path.replace('/', '.')}.{protocol_dir.name}.{sub_dir.name}"
                                        )
                                    except ImportError as e:
                                        self.logger.error(
                                            f"Failed to load plugin {sub_dir.name}: {e}"
                                        )
            else:
                for sub_dir in plugin_dir.iterdir():
                    if sub_dir.is_dir() and not sub_dir.name.startswith("__"):
                        plugin_file = sub_dir / f"{sub_dir.name}.py"
                        if plugin_file.exists():
                            self.logger.debug(f"Found plugin: {sub_dir.name}")
                            plugins.append(sub_dir.name)
                            try:
                                importlib.import_module(
                                    f"panther.plugins.{plugin_path.replace('/', '.')}.{sub_dir.name}"
                                )
                            except ImportError as e:
                                self.logger.error(
                                    f"Failed to load plugin {sub_dir.name}: {e}"
                                )
            all_plugins[plugin_type] = sorted(plugins)

        return all_plugins



from panther.core.results.result_handler import ResultHandler


class ResultCollector:
    """
    ResultCollector is responsible for collecting and processing results using registered handlers.
    Attributes:
        handlers (Dict[str, List[ResultHandler]]): A dictionary mapping result types to lists of result handlers.
    Methods:
        __init__():
            Initializes the ResultCollector with an empty handlers dictionary.
        collect(result: Dict) -> None:
            Processes the given result using the appropriate handlers based on the result type.
            Args:
                result (Dict): The result to be processed, which should contain a 'type' key to determine the handlers.
    """

    def __init__(self) -> None:
        self.handlers: dict[str, list[ResultHandler]] = {}

    def register_handler(self, result_type: str, handler: ResultHandler) -> None:
        """Registers a handler for a specific result type."""
        if result_type not in self.handlers:
            self.handlers[result_type] = []
        self.handlers[result_type].append(handler)

    def collect(self, result: dict) -> None:
        # TODO
        for handler in self.handlers.get(result.get("type"), []):
            handler.handle(result)



from dataclasses import dataclass, field
from enum import Enum


@dataclass
class Parameter:
    value: str | None = None
    description: str | None = None


@dataclass
class VersionBase:
    version: str
    commit: str
    dependencies: list[dict[str, str]]


# Implementation Configuration
# IUT: Implementation Under Test
# Tester: Implementation used for testing
ImplementationType = Enum("ImplementationType", ["iut", "testers"])


@dataclass
class ImplementationConfig:
    """
    ImplementationConfig class represents the configuration for an implementation.

    Attributes:
        name (str): The name of the implementation (e.g., picoquic, panther_ivy).
        type (ImplementationType): The type of implementation, must be either "iut" or "testers".
        shadow_compatible (bool): Indicates if the implementation is compatible with shadow. This field must be ignored by OmegaConf.
        gperf_compatible (bool): Indicates if the implementation is compatible with gperf.
    """

    name: str  # Implementation name (e.g., picoquic, panther_ivy)
    type: ImplementationType = (
        ImplementationType.iut
    )  # Must be either "iut" or "testers"
    shadow_compatible: bool = field(
        default=False
    )  # This field must be ignored by OmegaConf
    gperf_compatible: bool = field(default=False)



#!/usr/bin/env python3.9

import os
from flask import (
    Flask,
    redirect,
)
from flask_cors import CORS
from omegaconf import OmegaConf

from panther.config.config_global_schema import GlobalConfig
from panther.config.config_manager import ConfigLoader
from panther.core.experiment_manager import ExperimentManager


def create_app(config_loader: ConfigLoader, global_config: GlobalConfig, args):
    app = Flask(
        "panther_webapp",
        static_folder="panther/webapp/static/",
        template_folder="panther/webapp/templates/",
    )
    # # app.logger.setLevel(logging.DEBUG)
    print(f"Flask app template - {app.template_folder} - {os.getcwd()}")
    app.secret_key = "ElNiakDummyKey"
    app.config["SESSION_TYPE"] = "filesystem"
    app.config["SESSION_PERMANENT"] = False
    app.config["APPLICATION_ROOT"] = "panther/webapp/templates/"
    CORS(app, resources={r"/*": {"origins": "*"}})

    app.config["config_loader"] = config_loader
    app.config["global_config"] = global_config
    experiment_manager = ExperimentManager(
        global_config=global_config, experiment_name=args.experiment_name
    )
    experiment_manager.test_cases
    app.config["experiment_manager"] = experiment_manager

    experiment_config = config_loader.load_and_validate_experiment_config()
    print(f"Experiment Config: {OmegaConf.to_yaml(experiment_config)}")
    app.config["experiment_config"] = experiment_config
    # Once we have the experiments configurations, we can initialize the experiment
    experiment_manager.initialize_experiments(experiment_config)

    from .experiment_setup import exp_manager

    app.register_blueprint(exp_manager, url_prefix="/")
    app.logger.info(f"Flask app template - {app.template_folder}")

    @app.after_request
    def add_header(r):
        """
        It sets the cache control headers to prevent caching

        :param r: The response object
        :return: the response object with the headers added.
        """
        r.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
        r.headers["Pragma"] = "no-cache"
        r.headers["Expires"] = "0"
        r.headers["Cache-Control"] = "public, max-age=0"
        r.headers.add("Access-Control-Allow-Headers", "authorization,content-type")
        r.headers.add(
            "Access-Control-Allow-Methods",
            "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT",
        )
        r.headers.add("Access-Control-Allow-Origin", "*")
        return r

    @app.route("/")
    def redirection():
        """
        It redirects the user to the index.html page
        :return: a redirect to the index.html page.
        """
        return redirect("index", code=302)

    return app


def run(config_loader: ConfigLoader, global_config: GlobalConfig, args):
    print("Running webapp")
    app = create_app(
        config_loader=config_loader, global_config=global_config, args=args
    )
    app.run(host="0.0.0.0", port=8080, use_reloader=True, threaded=True, debug=True)




from dataclasses import dataclass
from omegaconf import MISSING

from panther.plugins.environments.config_schema import EnvironmentConfig

# Network Environment Configuration
@dataclass
class NetworkEnvironmentConfig(EnvironmentConfig):
    type: str = MISSING



from enum import Enum
from dataclasses import dataclass

# Logging Configuration
LoggingLevel = Enum("LoggingLevel", ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])


@dataclass
class LoggingConfig:
    """
    Configuration class for logging settings.

    Attributes:
        level (LoggingLevel): The logging level, with limited valid values.
        format (str): The format string for log messages.
    """

    level: LoggingLevel = LoggingLevel.DEBUG  # Limited valid values
    format: str = "%(asctime)s [%(levelname)s] - %(module)s - %(message)s"


# Paths Configuration
@dataclass
class PathsConfig:
    """
    PathsConfig is a configuration class that defines various directory paths used in the application.

    Attributes:
        output_dir (str): Directory path where output files are stored. Default is "panther/outputs".
        log_dir (str): Directory path where log files are stored. Default is "panther/outputs/logs".
        config_dir (str): Directory path where configuration files are stored. Default is "panther/configs".
        plugin_dir (str): Directory path where plugin files are stored. Default is "panther/plugins".
        services_dir (str): Directory path where service files are stored. Default is "services".
        iut_dir (str): Directory path where IUT (Implementation Under Test) files are stored. Default is "iut".
        testers_dir (str): Directory path where tester files are stored. Default is "testers".
    """

    output_dir: str = "panther/outputs"
    log_dir: str = "panther/outputs/logs"
    config_dir: str = "panther/configs"
    plugin_dir: str = "panther/plugins"
    services_dir: str = "services"
    iut_dir: str = "iut"
    testers_dir: str = "testers"


@dataclass
class AdditionalPathsConfig:
    """
    AdditionalPathsConfig is a configuration class that holds directory paths for various environments.

    Attributes:
        exec_env_dir (str): Path to the execution environment directory.
        net_env_dir (str): Path to the network environment directory.
        iut_dir (str): Path to the IUT (Implementation Under Test) directory.
        testers_dir (str): Path to the testers directory.
    """

    exec_env_dir: str = ""
    net_env_dir: str = ""
    iut_dir: str = ""
    testers_dir: str = ""


# Docker Configuration
@dataclass
class DockerConfig:
    """
    Configuration settings for Docker operations.

    Attributes:
        build_docker_image (bool): Flag to determine if the Docker image should be built.
        remove_docker_image (bool): Flag to determine if the Docker image should be removed.
        remove_docker_container (bool): Flag to determine if the Docker container should be removed.
        remove_docker_network (bool): Flag to determine if the Docker network should be removed.
        remove_docker_volume (bool): Flag to determine if the Docker volume should be removed.
    """

    build_docker_image: bool = True
    remove_docker_image: bool = True
    remove_docker_container: bool = True
    remove_docker_network: bool = True
    remove_docker_volume: bool = True


# Feature Configuration
@dataclass
class FeatureConfig:
    """
    FeatureConfig class is used to configure global features for the application.

    Attributes:
        logger_observer (bool): Indicates whether the logger observer feature is enabled. Default is True.
        storage_handler (bool): Indicates whether the storage handler feature is enabled. Default is True.
    """

    logger_observer: bool = True
    storage_handler: bool = True


@dataclass
class GlobalConfig:
    logging: LoggingConfig = LoggingConfig()
    paths: PathsConfig = PathsConfig()
    optional_paths: AdditionalPathsConfig = AdditionalPathsConfig()
    docker: DockerConfig = DockerConfig()
    features: FeatureConfig = FeatureConfig()



from dataclasses import dataclass, field
from typing import List, Optional

from panther.plugins.environments.execution_environment.config_schema import ExecutionEnvironmentConfig


@dataclass
class StraceConfig(ExecutionEnvironmentConfig):
    """
    StraceConfig is a configuration class for setting up and running strace in a specific execution environment.

    Attributes:
        strace_binary (str): Path to the strace binary. Default is "/usr/bin/strace".
        excluded_syscalls (List[str]): List of syscalls to exclude from tracing. Default includes various time-related syscalls.
        include_kernel_stack (bool): Whether to include the kernel stack in the trace output. Default is True.
        trace_network_syscalls (bool): Whether to focus on network-related syscalls (e.g., connect, send, recv). Default is True.
        timeout (Optional[int]): Timeout for strace execution in seconds. Default is 60 seconds.
        output_file (str): Path to the strace log output file. Default is "/app/logs/strace.log".
        additional_parameters (List[str]): Additional parameters to pass to strace. Default is an empty list.
        monitored_process (Optional[str]): Name of the process to monitor, if not using PID-based monitoring. Default is None.
        network_focus (bool): Indicates if strace should emphasize network protocol syscalls. Default is True.
    """
    strace_binary: str = "/usr/bin/strace"  # Path to the strace binary
    excluded_syscalls: List[str] = field(
        default_factory=lambda: [
            "nanosleep",
            "getitimer",
            "alarm",
            "setitimer",
            "gettimeofday",
            "times",
            "rt_sigtimedwait",
            "utime",
            "adjtimex",
            "settimeofday",
            "time",
        ]
    )  # List of syscalls to exclude
    include_kernel_stack: bool = True  # Include kernel stack in the trace output
    trace_network_syscalls: bool = True  # Focus on network-related syscalls (connect, send, recv, etc.)
    timeout: Optional[int] = 60  # Timeout for strace execution in seconds
    output_file: str = "/app/logs/strace.log"  # Path to the strace log output
    additional_parameters: List[str] = field(default_factory=list)  # Additional parameters for strace
    monitored_process: Optional[str] = None  # Process name to monitor (if not PID-based)
    network_focus: bool = True  # Indicate if strace should emphasize network protocol syscalls




import traceback
from abc import abstractmethod
import os

from jinja2 import Environment, FileSystemLoader
from omegaconf import OmegaConf

from panther.plugins.services.services_interface import IServiceManager

from panther.config.config_experiment_schema import TestConfig

from panther.config.config_global_schema import GlobalConfig

from panther.plugins.plugin_loader import PluginLoader

from panther.core.observer.event_manager import EventManager
from panther.plugins.environments.config_schema import EnvironmentConfig
from panther.plugins.environments.execution_environment.execution_environment_interface import (
    IExecutionEnvironment,
)
from panther.plugins.environments.environment_interface import IEnvironmentPlugin


class INetworkEnvironment(IEnvironmentPlugin):
    """
    INetworkEnvironment is an abstract class that defines the interface for network environment plugins.
    It inherits from IEnvironmentPlugin and provides methods for setting up, updating, and managing network environments.
    Attributes:
        docker_name (str): Name of the Docker container.
        execution_environment (list): List of execution environments.
        network_name (str): Name of the network.
        execution_environments (list): List of execution environments.
        services (dict): Dictionary of services.
        deployment_commands (dict): Dictionary of deployment commands.
        timeout (int): Timeout value.
        global_config (GlobalConfig): Global configuration.
        test_config (TestConfig): Test configuration.
        services_managers (list): List of service managers.
        logger (Logger): Logger instance.
        jinja_env (Environment): Jinja2 environment for template rendering.
    Methods:
        __init__(env_config_to_test, output_dir, env_type, env_sub_type, event_manager):
            Initializes the network environment with the given configuration.
        setup_execution_plugins(timestamp):
            Sets up execution plugins for the environment.
        update_environment(execution_environment, global_config, plugin_loader, services_managers, test_config):
            Updates the environment with the given configuration and services.
        create_log_dir(service):
            Creates a log directory for the given service.
        generate_from_template(template_name, paths, timestamp, rendered_out_file, out_file, additional_param=None):
            Generates a file from a Jinja2 template.
        get_docker_name():
            Retrieves the Docker container name.
        resolve_environment_variables(env_vars):
            Resolves environment variables incrementally, ensuring no duplication and preserving unresolved tokens.
        is_network_environment():
        generate_environment_services(paths, timestamp):
            Abstract method to generate the services required for the network environment.
        prepare_environment():
            Abstract method to prepare the environment for running experiments.
        launch_environment_services():
            Abstract method to launch the services in the network environment.
        deploy_services():
            Abstract method to deploy the specified services in the network environment.
        setup_environment(services_managers, test_config, global_config, timestamp, plugin_loader, execution_environment):
            Abstract method to set up the required environment before running experiments.
        teardown_environment():
            Abstract method to tear down the environment after experiments are completed.
    """

    def __init__(
        self,
        env_config_to_test: EnvironmentConfig,
        output_dir: str,
        env_type: str,
        env_sub_type: str,
        event_manager: EventManager,
    ):
        super().__init__(
            env_config_to_test, output_dir, env_type, env_sub_type, event_manager
        )
        self.docker_name = None
        self.execution_environment = None
        self.network_name = f"{env_sub_type}_network"
        self.execution_environments = []

        self.services = {}
        self.deployment_commands = {}
        self.timeout = 60

        self.global_config = None
        self.test_config = None
        self.services_managers = None

        self.logger.debug(
            f"Environment settings: {self.env_config_to_test} in {self.templates_dir}"
        )
        self.jinja_env = Environment(loader=FileSystemLoader(self.templates_dir))
        self.jinja_env.filters["realpath"] = lambda x: os.path.abspath(x)
        self.jinja_env.filters["is_dict"] = lambda x: isinstance(x, dict)
        self.jinja_env.trim_blocks = True
        self.jinja_env.lstrip_blocks = True

    def setup_execution_plugins(self, timestamp):
        """
        Sets up the execution plugins for each execution environment.

        This method iterates through the list of execution environments and sets up each one by calling its
        `setup_environment` method with the necessary parameters. If an error occurs during the setup of any
        execution environment, it logs the error along with the traceback.

        Args:
            timestamp (str): The timestamp to be used during the setup of the execution environments.

        Raises:
            Exception: If an error occurs during the setup of any execution environment, it is caught and logged.
        """
        for execution_env in self.execution_environment:
            try:
                self.logger.debug(f"Setting up execution environment: {execution_env}")
                execution_env.setup_environment(
                    services_managers=self.services_managers,
                    test_config=self.test_config,
                    global_config=self.global_config,
                    timestamp=timestamp,
                    plugin_loader=self.plugin_loader,
                )
            except Exception as e:
                self.logger.error(
                    f"Failed to setup execution environment: {e}\n{traceback.format_exc()}"
                )

    def update_environment(
        self,
        execution_environment,
        global_config,
        plugin_loader,
        services_managers,
        test_config,
    ):
        """
        Updates the network environment with the provided configuration and services.

        Args:
            execution_environment (Any): The execution environment to be used.
            global_config (OmegaConf): The global configuration settings.
            plugin_loader (Any): The plugin loader instance.
            services_managers (List[IServiceManager]): A list of service manager instances.
            test_config (OmegaConf): The test configuration settings.

        Returns:
            None
        """
        self.services_managers: list[IServiceManager] = services_managers
        self.test_config = test_config
        self.execution_environment = execution_environment
        self.plugin_loader = plugin_loader
        self.global_config = global_config
        self.logger.debug("Setup environment with:")
        for service in self.services_managers:
            self.logger.debug(f"Service: {service}")
        self.logger.debug(f"Test Config: {OmegaConf.to_yaml(self.test_config)}")
        self.logger.debug(f"Global Config: {OmegaConf.to_yaml(self.global_config)}")

    def create_log_dir(self, service: IServiceManager):
        """
        Creates a log directory for the given service if it does not already exist.

        Args:
            service (IServiceManager): The service manager instance containing the service name.

        Logs:
            Info: Logs the creation of the log directory if it was created.
        """
        log_dir = os.path.join(self.log_dirs, service.service_name)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
            self.logger.info(f"Created log directory: {log_dir}")

    def generate_from_template(
        self,
        template_name,
        paths,
        timestamp,
        rendered_out_file,
        out_file,
        additional_param=None,
    ):
        """
        Generates a configuration file from a Jinja2 template.

        Args:
            template_name (str): The name of the Jinja2 template to use.
            paths (dict): A dictionary of paths to be used in the template.
            timestamp (str): A timestamp to be included in the rendered template.
            rendered_out_file (str): The file path where the rendered template will be saved.
            out_file (str): The file path where the final output will be saved.
            additional_param (dict, optional): Additional parameters to be passed to the template. Defaults to None.

        Returns:
            None
        """
        template = self.jinja_env.get_template(template_name)
        self.logger.debug(f"Template: {template}")
        self.logger.debug(f"Services: {self.services_managers}")
        self.logger.debug(f"Deployment Info: {self.test_config}")
        rendered = template.render(
            services=self.services_managers,
            test_config=self.test_config,
            paths=paths,
            timestamp=timestamp,
            additional_param=additional_param,
            log_dir=self.log_dirs,
            experiment_name=self.output_dir.split("/")[-1],
        )
        # Write the rendered content to shadow.generated.yml
        with open(out_file, "w") as f:
            f.write(rendered)
        with open(rendered_out_file, "w") as f:
            f.write(rendered)

    def get_docker_name(self):
        """
        Retrieves and sets the Docker container name by building a Docker image from a specified path.

        This method uses the `plugin_loader` to build a Docker image from the provided Dockerfile path,
        Docker name, and Docker version. It then extracts and sets the Docker container name by splitting
        the resulting Docker image name at the colon (':') character.

        Returns:
            str: The name of the Docker container.
        """
        self.docker_name = self.plugin_loader.build_docker_image_from_path(
            self.services_network_docker_file_path,
            self.docker_name,
            self.docker_version,
        )
        self.docker_name = self.docker_name.split(":")[0]

    def resolve_environment_variables(self, env_vars):
        """
        Resolves environment variables incrementally, ensuring no duplication
        and preserving unresolved tokens. Processes variables in dependency order.

        :param env_vars: dict, environment variables with potential references.
        :return: dict, resolved environment variables.
        """
        resolved_env = {}

        self.logger.debug("Initial environment variables:")
        for k, v in env_vars.items():
            self.logger.debug(f"{k}: {v}")

        for key, value in env_vars.items():
            if isinstance(value, str):
                resolved_value = value
                self.logger.debug(
                    f"Resolving variable: {key} - Original value: {value}"
                )
                for (
                    var_name,
                    var_value,
                ) in resolved_env.items():  # Use already resolved variables
                    if (
                        f"${{{var_name}}}" in resolved_value
                        or f"${var_name}" in resolved_value
                    ):
                        resolved_value = resolved_value.replace(
                            f"${{{var_name}}}", var_value
                        )
                        resolved_value = resolved_value.replace(
                            f"${var_name}", var_value
                        )
                        self.logger.debug(
                            f"Replaced ${var_name} in {key} with {var_value}"
                        )
                resolved_value = resolved_value.replace("$", "$$")
                resolved_env[key] = resolved_value

        self.logger.debug("Final resolved environment variables without duplication:")
        for k, v in resolved_env.items():
            self.logger.debug(f"{k}: {v}")

        return resolved_env

    def is_network_environment(self):
        """
        Returns True if the plugin is a network environment.
        """
        return True

    @abstractmethod
    def generate_environment_services(self, paths: dict[str, str], timestamp: str):
        """
        Generates the services required for the network environment.

        :param services: A dictionary containing the services to be generated.
        :return: A list of generated services.
        """
        pass

    @abstractmethod
    def prepare_environment(self):
        """
        Prepares the environment for running experiments.
        """
        pass

    @abstractmethod
    def launch_environment_services(self):
        """
        Launches the services in the network environment.
        """
        pass

    @abstractmethod
    def deploy_services(self):
        """
        Deploys the specified services in the network environment.
        """
        pass

    @abstractmethod
    def setup_environment(
        self,
        services_managers: list[IServiceManager],
        test_config: TestConfig,
        global_config: GlobalConfig,
        timestamp: str,
        plugin_loader: PluginLoader,
        execution_environment: list[IExecutionEnvironment],
    ):
        """
        Sets up the required environment before running experiments.
        """
        raise NotImplementedError()

    @abstractmethod
    def teardown_environment(self):
        """
        Tears down the environment after experiments are completed.
        """
        pass



from dataclasses import dataclass
from enum import Enum

from panther.plugins.protocols.config_schema import (
    ProtocolConfig,
    RoleEnum,
    ProtocolType,
)


# TODO manage versions with specificities (e.g draft27, draft27-vuln1, ...)
VersionEnum = Enum("VersionEnum", ["rfc9000", "draft29", "draft27"])


@dataclass
class QuicConfig(ProtocolConfig):
    """
    QuicConfig is a configuration class for the QUIC protocol.

    Attributes:
        name (str): The name of the protocol, default is "QUIC".
        version (VersionEnum): The version of the protocol, default is VersionEnum.rfc9000.
        role (RoleEnum): The role of the protocol, either server or client, default is RoleEnum.server.
        target (str | None): An optional target service name, default is None.
        protocol_type (ProtocolType): The type of protocol, default is ProtocolType.client_server.
    """
    name: str = "QUIC"
    version: VersionEnum = VersionEnum.rfc9000  # Protocol version (e.g., rfc9000)
    role: RoleEnum = RoleEnum.server  # Role (server or client)
    target: str | None = None  # Optional target service name
    protocol_type: ProtocolType = ProtocolType.client_server



from abc import ABC
from typing import Any


class Event(ABC):
    def __init__(self, name: str, data: dict[str, Any] = None):
        self.name = name
        self.data = data or {}

    def __str__(self):
        return f"Event(name='{self.name}', data={self.data})"

    def __repr__(self):
        return str(self)



from dataclasses import dataclass
from enum import Enum


# Base protocol configuration
@dataclass
class ProtocolBase:
    name: str

    def validate(self):
        """Common validation logic for protocols."""
        pass


ProtocolType = Enum("ProtocolType", ["peer_to_peer", "client_server"])
RoleEnum = Enum("RoleEnum", ["server", "client", "peer"])


@dataclass
class ProtocolConfig:
    name: str | None = None
    version: str | None = None
    role: str | None = None
    target: str | None = None
    protocol_type: ProtocolType = ProtocolType.client_server
